# This file is part of Minnt <http://github.com/foxik/minnt/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Any, Self

import torch

from ..logger import Logger
from ..type_aliases import AnyArray


class BaseLogger(Logger):
    """An abstract logger providing base functionality for other loggers."""

    def __del__(self) -> None:
        """Ensure proper resource cleanup on deletion."""
        self.close()

    def log_figure(self, label: str, figure: Any, epoch: int, tight_layout: bool = True, close: bool = True) -> Self:
        import matplotlib.pyplot as plt
        import matplotlib.backends.backend_agg as plt_backend_agg

        tight_layout and figure.tight_layout()
        canvas = plt_backend_agg.FigureCanvasAgg(figure)
        canvas.draw()
        width, height = figure.canvas.get_width_height()
        image = torch.frombuffer(canvas.buffer_rgba(), dtype=torch.uint8).view(height, width, 4)
        close and plt.close(figure)

        return self.log_image(label, image, epoch)

    @staticmethod
    def format_epoch_logs(
        logs: dict[str, float], epoch: int, epochs: int | None = None, elapsed: float | None = None,
    ) -> str:
        return " ".join(
            [f"Epoch {epoch}" + (f"/{epochs}" if epochs is not None else "")]
            + ([f"{elapsed:.1f}s"] if elapsed is not None else [])
            + [f"{k}={v:#.{0 < abs(v) < 2e-4 and '2e' or '4f'}}" for k, v in logs.items()]
        )

    @staticmethod
    def preprocess_audio(audio: AnyArray) -> torch.Tensor:
        """Produce a CPU-based [torch.Tensor][] with `dtype=torch.int16` and shape `(L, {1/2})`."""
        audio = torch.as_tensor(audio, device="cpu")
        audio = audio * 32_767 if audio.dtype.is_floating_point else audio
        audio = audio.clamp(-32_768, 32_767).to(torch.int16)
        assert audio.ndim == 1 or (audio.ndim == 2 and audio.shape[1] in (1, 2)), \
            "Audio must have shape (L,) or (L, 1/2)"
        if audio.ndim == 1:
            audio = audio.unsqueeze(-1)
        return audio

    @staticmethod
    def preprocess_image(image: AnyArray) -> torch.Tensor:
        """Produce a CPU-based [torch.Tensor][] with `dtype=torch.uint8` and shape `(H, W, {1/3/4})`."""
        image = torch.as_tensor(image, device="cpu")
        image = (image * 255 if image.dtype.is_floating_point else image).clamp(0, 255).to(torch.uint8)
        assert image.ndim == 2 or (image.ndim == 3 and image.shape[2] in (1, 2, 3, 4)), \
            "Image must have shape (H, W) or (H, W, 1/2/3/4)"
        if image.ndim == 2:
            image = image.unsqueeze(-1)
        if image.shape[2] == 2:
            # Convert to RGBA
            image = torch.stack([image[:, :, 0]] * 3 + [image[:, :, 1]], dim=-1)
        return image



# This file is part of Minnt <http://github.com/foxik/minnt/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from abc import ABC, abstractmethod

import torch

from .type_aliases import TensorOrTensors


class Metric(torch.nn.Module, ABC):
    """An abstract metric interface."""

    @abstractmethod
    def update(
        self, y: TensorOrTensors, y_true: TensorOrTensors | None = None, sample_weights: TensorOrTensors | None = None,
    ) -> None:
        """Update the internal state of the metric with new predictions and possibly gold targets.

        Optional sample weights might be provided if supported by the metric.

        Parameters:
          y: The predicted outputs.
          y_true: Optional ground-truth targets.
          sample_weights: Optional sample weights.
        """
        ...

    @abstractmethod
    def compute(self) -> torch.Tensor:
        """Compute the accumulated metric value.

        Returns:
          A (usually scalar) tensor representing the accumulated metric value.
        """
        ...

    @abstractmethod
    def reset(self) -> None:
        """Reset the internal state of the metric."""
        ...



# This file is part of Minnt <http://github.com/foxik/minnt/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Protocol

import torch

from .type_aliases import TensorOrTensors


class Loss(Protocol):
    """An abstract loss function interface."""

    def __call__(
        self, y: TensorOrTensors, y_true: TensorOrTensors, sample_weights: TensorOrTensors | None = None,
    ) -> torch.Tensor:
        """Compute loss of the given predictions and gold targets, optionally with sample weights.

        Parameters:
          y: The predicted outputs.
          y_true: The ground-truth targets.
          sample_weights: Optional sample weights.

        Returns:
          A tensor representing the computed loss.
        """
        ...



# This file is part of Minnt <http://github.com/foxik/minnt/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from .base_logger import BaseLogger
from .filesystem_logger import FileSystemLogger
from .multi_logger import MultiLogger
from .tensorboard_logger import TensorBoardLogger
from .wandb_logger import WandbLogger



# This file is part of Minnt <http://github.com/foxik/minnt/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# TrainableModule
from .trainable_module import TrainableModule

# TransformedDataset
from .transformed_dataset import TransformedDataset

# Type aliases
from .type_aliases import AnyArray, HasCompute, Logs, Reduction, Tensor, TensorOrTensors

# Vocabulary
from .vocabulary import Vocabulary

# Utils
from .initializers_override import global_keras_initializers
from .startup import startup
from .version import __version__

# Callbacks
from .callback import Callback, STOP_TRAINING
from . import callbacks

# Loggers
from .logger import Logger
from . import loggers

# Losses
from .loss import Loss
from . import losses

# Metrics
from .metric import Metric
from . import metrics



# This file is part of Minnt <http://github.com/foxik/minnt/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from collections.abc import Iterable


class Vocabulary:
    """ A class for managing mapping between strings and indices.

    The vocabulary is initialized with a list of strings.

    It always contains a special padding token [Vocabulary.PAD][minnt.Vocabulary.PAD]
    at index 0, and optionally an unknown token [Vocabulary.UNK][minnt.Vocabulary.UNK]
    at index 1 (when [Vocabulary.has_unk][minnt.Vocabulary.has_unk]).
    """
    PAD: int = 0
    """The index of the padding token."""
    UNK: int = 1
    """The index of the unknown token, which might not be present."""

    def __init__(self, strings: Iterable[str], add_unk: bool = False) -> None:
        """Initialize the vocabulary with the given list of strings.

        The [Vocabulary.PAD][minnt.Vocabulary.PAD] is always the first token in the vocabulary;
        [Vocabulary.UNK][minnt.Vocabulary.UNK] is the second token but only when `add_unk=True`.
        """
        self._strings = ["[PAD]"] + (["[UNK]"] if add_unk else [])
        self._strings.extend(strings)
        self._string_map = {string: index for index, string in enumerate(self._strings)}
        self._has_unk = add_unk

    @property
    def has_unk(self) -> bool:
        """A boolean property indicating whether the vocabulary was constructed with an UNK token."""
        return self._has_unk

    def __len__(self) -> int:
        """The number of strings in the vocabulary.

        Returns:
          The size of the vocabulary.
        """
        return len(self._strings)

    def __iter__(self) -> Iterable[str]:
        """Return an iterator over strings in the vocabulary.

        Returns:
          An iterator over strings in the vocabulary.
        """
        return iter(self._strings)

    def string(self, index: int) -> str:
        """Convert vocabulary index to string.

        Parameters:
          index: The vocabulary index.

        Returns:
          The string corresponding to the given index.
        """
        return self._strings[index]

    def strings(self, indices: Iterable[int]) -> list[str]:
        """Convert a sequence of vocabulary indices to strings.

        Parameters:
          indices: An iterable of vocabulary indices.

        Returns:
          A list of strings corresponding to the given indices.
        """
        return [self._strings[index] for index in indices]

    def index(self, string: str) -> int | None:
        """Convert string to vocabulary index.

        Parameters:
          string: The string to convert.

        Returns:
          The index corresponding to the given string. If the string is not found in the vocabulary, then

            - if the vocabulary was constructed with an UNK token, it returns [Vocabulary.UNK][minnt.Vocabulary.UNK];
            - otherwise, it returns `None`.
        """
        return self._string_map.get(string, self.UNK if self._has_unk else None)

    def indices(self, strings: Iterable[str]) -> list[int | None]:
        """Convert a sequence of strings to vocabulary indices.

        Parameters:
          strings: An iterable of strings to convert.

        Returns:
          A list of indices corresponding to the given strings. For each string not found in the vocabulary, it returns

            - [Vocabulary.UNK][minnt.Vocabulary.UNK] if the vocabulary was constructed with an UNK token;
            - otherwise, it returns `None`.
        """
        return [self._string_map.get(string, self.UNK if self._has_unk else None) for string in strings]
